#include <iostream>
#include <fstream>
#include <vector>

// texgen.cpp: Generates textures for amazing
// (C) 2003 Jeremy Stanley

// This program is licensed under the GNU General Public License.
// See COPYING for details.

// Takes raw binary input (such as generated by Matthew Johnson's iStudio)
// Must be 128x64 pixels in size, 2 planes (four grayscales)

using namespace std;

void convert(unsigned char *dest, const unsigned char *src);

int main(int argc, char **argv)
{
	if (argc < 3)
	{
		cerr << "Usage: " << argv[0] << 
			" input.bin output.c\n";
		
		return 1;
	}

	unsigned char in_buf[2048], out_buf[2048];

	ifstream in_bin(argv[1], ios_base::binary);
	if ( !in_bin )
	{
		cerr << "Couldn't open input file " << argv[1] << endl;
		return 1;
	}
	in_bin.seekg(0, ios_base::end);
	if (2048 != in_bin.tellg())
	{
		cerr << "Incorrect image size" << endl;
		return 1;
	}
	in_bin.seekg(0, ios_base::beg);
	in_bin.read(reinterpret_cast<char *>(&in_buf[0]), 2048);
	in_bin.close();

	convert(&out_buf[0], &in_buf[0]); // light plane
	convert(&out_buf[1024], &in_buf[1024]); // dark plane

	ofstream out_c(argv[2]);
	out_c << "unsigned char texture[] = {" << hex << endl;
	for(int i = 0; i < 2048; i++)
	{
		out_c << "0x" << (int)out_buf[i];

		if (i < 2047)
			out_c << ", ";

		if (0 == ((i + 1) % 8))
			out_c << endl;
	}
	out_c << "}; " << endl << dec;
	out_c.close();

	return 0;
}

// prepares the 128x64 texture for use in amazing
// (i.e., changes from row-major order to column-major order
// for more efficient texturing)
void convert(unsigned char *dest, const unsigned char *src)
{
	for(int x = 0; x < 128; x++)
	{
		for(int y = 0; y < 64; y++)
		{
			unsigned char *dp = dest + x * (64/8) + (y/8);
			unsigned char dest_mask = (0x80 >> (y%8));
			if (*(src + y * (128/8) + (x/8)) & (0x80 >> (x%8)))
			{
				*dp |= dest_mask;
			}
			else
			{
				*dp &= ~dest_mask;
			}
		}
	}
}