// (C) 2003 by Jeremy Stanley
// This program is licensed under the GNU General Public License.
// See COPYING for details.

// implementation of world.c

#include "world.h"
#include "fix.h"
#include "ang.h"
#include "graphics.h"
#include <tigcclib.h>

#ifdef TEXTURE_WALLS

extern unsigned char stonetex[];

#endif

#define ABS(x) (((x) < 0) ? -(x) : (x))

unsigned char *map = NULL;
short m_width = 0, m_height = 0;
FIX m_fixwidth = 0, m_fixheight = 0;

// Prototypes for non-exported functions
short openthis();
void world_burrowmaze(short x, short y);

// Openness factor for random mazes
#define OPENNESS 20

// Use this function to set up an external map
void world_setmap(unsigned char *map_ptr, short width, short height)
{
	map = map_ptr;
	m_width = width;
	m_height = height;
	m_fixwidth = INT_TO_FIX(m_width); 
	m_fixheight = INT_TO_FIX(m_height); 
}

// Use this function to allow the raycaster to create a random maze
// It must be deleted by calling world_delrandmap() before the program ends
short world_setrandmap(unsigned char *mapbuf, short width, short height, short walltex)
{
	int bytes = width * height;
	short xx, xy;

	m_width = width;
	m_height = height;
	m_fixwidth = INT_TO_FIX(m_width);
	m_fixheight = INT_TO_FIX(m_height);

  map = mapbuf;

	// Fill the world solid
	memset(map, walltex, bytes);

	// Bore out a maze
	world_burrowmaze(1, 1);
	
	// Place an exit [texture 2] somewhere accessible along the outer edge
	xx = rand() % width;
	xy = rand() % height;
  switch( rand() & 3 ) {
     case 0: xx = 0; 
        while(get(1, xy)) { xy++; if(xy >= height) xy = 0; }
        break;
     case 1: xx = width - 1; 
        while(get(width - 2, xy)) { xy++; if(xy >= height) xy = 0; }
        break;
     case 2: xy = 0; 
        while(get(xx, 1)) { xx++; if(xx >= width) xx = 0; }
        break;
     case 3: xy = height - 1; 
        while(get(xx, width - 2)) { xx++; if(xx >= width) xx = 0; }
        break;
  }
  set(xx, xy, 2);
  
	return 1;
}

// Helper functions for getting and setting textures in the maze
short get(short x, short y)
{
	return map[y * m_width + x];
}

void set(short x, short y, short tex)
{
	map[y * m_width + x] = (unsigned char)tex;
}

// essentially a weighted coin world_burrowmaze uses to decide whether to make a redundant tunnel
short openthis()
{
    return (rand() % OPENNESS) == 0;
}

// Recursive component of random maze generator
void world_burrowmaze(short x, short y)
{
    int pos, nswp, i, j, arr[4] = {0, 1, 2, 3};

    // check if we've hit a boundary
    if(x < 1 || y < 1 || x >= (m_width-1) || y >= (m_height-1))
        return;

    // check if we've been here before
    if(get(x,y)==0)
        return;

    // burrow out this square
    set(x, y, 0);

    // burrow neighbors in random order
    for(nswp = 0; nswp < 8; nswp++)
    {
        i = rand() & 3;
        j = rand() & 3;

        pos = arr[i];
        arr[i] = arr[j];
        arr[j] = pos;
    }

    for(pos=0; pos<4; pos++)
    {
        switch(arr[pos])
        {
        case 0:   // burrow right
            if(x < (m_width-2)) {
                if(get(x+2, y) || openthis()) {
                    set(x+1, y, 0);
                    world_burrowmaze(x+2, y);
                }
            }
            break;
        case 1:    // burrow left
            if(x > 1) {
                if(get(x-2, y) || openthis()) {
                    set(x-1, y, 0);
                    world_burrowmaze(x-2, y);
                }
            }
            break;
        case 2:    // burrow up
            if(y > 1) {
                if(get(x, y-2) || openthis()) {
                    set(x, y-1, 0);
                    world_burrowmaze(x, y-2);
                }
            }
            break;
        case 3:    // burrow down
            if(y < (m_height-2)) {
                if(get(x, y+2) || openthis()) {
                    set(x, y+1, 0);
                    world_burrowmaze(x, y+2);
                }
            }
            break;
        }
    }
}

// Draws the map (2D) at the given coordinates
void world_drawmap(short u, short v)
{
	short i, j, p, q;

	for(i = 0; i < m_height; i++)
	{
		for(j = 0; j < m_width; j++)
		{
			if(get(j, i) == 1)
			{
			  p = u + (j << 1); 
			  q = v + (i << 1);
			  
        pixelcol89(p, q + 1, 1, GetPlane(LIGHT_PLANE), 30);
        pixelcol89(p + 1, q, 2, GetPlane(LIGHT_PLANE), 30);
        pixelcol89(p, q, 2, GetPlane(DARK_PLANE), 30);
        pixelcol89(p + 1, q, 2, GetPlane(DARK_PLANE), 30);
			}
		}
	}
}

// direct port of amazing's routine, converted to fixed-point math
void world_render3D(FIX x, FIX y, short vangle, void *buf)
{
    short a, h, th, top;
    FIX x0, dx, y0, dy;
    FIX dst_x, dst_y;
    FIX tx_x, tx_y;
    short ix, iy, dix, diy;
    short tex_x, tex_y;
    short col, a1, angle, va;
    short xplot, yplot;
#ifdef TEXTURE_WALLS
    short txc;
#endif

    for(col=0, a1=vangle+WID/2; col < WID; col++, a1--)
    {
        angle = a1;
        if(angle >= ANG_360) angle -= ANG_360;
        else if(angle < 0) angle += ANG_360;

        va = ABS(a1 - vangle);

        dst_x = -1;
        dst_y = -1;

        if(angle==0)          // straight right
        {
            iy = FIX_TO_INT(y);
            for(ix = FIX_TO_INT(x) + 1; (tex_y=get(ix, iy))==0; ix++);
            dst_y = FIX_MUL((INT_TO_FIX(ix) - x), cos_t[va]);
            tx_y = FIX_FPART(y);
            goto plot;
        }
        else if(angle < ANG_90)   // first quadrant
        {
            a = angle;

            x0 = x + FIX_MUL(FIX_FPART(y), cot_t[a]);
            dx = cot_t[a];
            iy = FIX_TO_INT(y) - 1;
            diy = -1;

            y0 = y - FIX_MUL((FIX_1 - FIX_FPART(x)), tan_t[a]);
            dy = -tan_t[a];
            ix = FIX_TO_INT(x) + 1;
            dix = 1;
        }
        else if(angle == ANG_90)  // straight up
        {
            ix = FIX_TO_INT(x);
            for(iy = FIX_TO_INT(y) - 1; (tex_x=get(ix, iy))==0; iy--);
            dst_x = FIX_MUL((y - INT_TO_FIX(iy+1)), cos_t[va]);
            tx_x = FIX_FPART(x);
            goto plot;
        }
        else if(angle < ANG_180)  // second quadrant
        {
            a = ANG_180 - angle;

            x0 = x - FIX_MUL(FIX_FPART(y), cot_t[a]);
            dx = -cot_t[a];
            iy = FIX_TO_INT(y) - 1;
            diy = -1;

            y0 = y - FIX_MUL(FIX_FPART(x), tan_t[a]);
            dy = -tan_t[a];
            ix = FIX_TO_INT(x) - 1;
            dix = -1;
        }
        else if(angle == ANG_180) // straight left
        {
            iy = FIX_TO_INT(y);
            for(ix = FIX_TO_INT(x) - 1; (tex_y=get(ix, iy))==0; ix--);
            dst_y = FIX_MUL((x - INT_TO_FIX(ix+1)), cos_t[va]);
            tx_y = FIX_1 - FIX_FPART(y);
            goto plot;
        }
        else if(angle < ANG_270)  // third quadrant
        {
            a = angle - ANG_180;

            x0 = x - FIX_MUL((FIX_1 - FIX_FPART(y)), cot_t[a]);
            dx = -cot_t[a];
            iy = 1 + FIX_TO_INT(y);
            diy = 1;

            y0 = y + FIX_MUL(FIX_FPART(x), tan_t[a]);
            dy = tan_t[a];
            ix = FIX_TO_INT(x) - 1;
            dix = -1;
        }
        else if(angle == ANG_270) // straight down
        {
            ix = FIX_TO_INT(x);
            for(iy = FIX_TO_INT(y) + 1; (tex_x=get(ix, iy))==0; iy++);
            dst_x = FIX_MUL((INT_TO_FIX(iy) - y), cos_t[va]);
            tx_x = FIX_1 - FIX_FPART(x);
            goto plot;
        }
        else                   // fourth quadrant
        {
            a = ANG_360 - angle;

            x0 = x + FIX_MUL((FIX_1 - FIX_FPART(y)), cot_t[a]);
            dx = cot_t[a];
            iy = 1 + FIX_TO_INT(y);
            diy = 1;

            y0 = y + FIX_MUL((FIX_1 - FIX_FPART(x)), tan_t[a]);
            dy = tan_t[a];
            ix = 1 + FIX_TO_INT(x);
            dix = 1;
        }

        // trace col-intercepts
        if(x0 < m_fixwidth && x0 >= 0) {
            while((tex_x=get(FIX_TO_INT(x0), iy))==0) {
                iy += diy;
                x0 += dx;
                if(x0 >= m_fixwidth || x0 < 0)
                    goto next;
            }
            dst_x = FIX_DIV(FIX_MUL(ABS(x0 - x), cos_t[va]), cos_t[a]);
            tx_x = (diy < 0 ) ? FIX_FPART(x0) : FIX_1 - FIX_FPART(x0);
        }

next:
        // trace row-intercepts
        if(y0 < m_fixheight && y0 >= 0) {
            while((tex_y=get(ix, FIX_TO_INT(y0)))==0) {
                ix += dix;
                y0 += dy;
                if(y0 >= m_fixheight || y0 < 0)
                    goto plot;
            }
            dst_y = FIX_DIV(FIX_MUL(ABS(y0 - y), cos_t[va]), cos_t[ANG_Q1 - a]);	// cos_t[ANG_Q1 - a] = sin(a) since a < ANG_90
            tx_y = (dix > 0) ? FIX_FPART(y0) : FIX_1 - FIX_FPART(y0);
        }

plot:
        // determine which intersection was closer 
				// and do some zany fiddling to make sure the exit square isn't rendered :)
				xplot = (dst_x > 0) && (tex_x != 2) && ((tex_y == 2) || (dst_x < dst_y || dst_y <= 0));
				yplot = (dst_y > 0) && (tex_y != 2);

				// draw the texture column
        if(xplot) {
            th = FIX_TO_INT(FIX_DIV(REFHEIGHT, dst_x));
	          h = (th < HEI) ? th : HEI;
      			top = HEI/2 - (h>>1);
#ifdef TEXTURE_WALLS
      			txc = tx_x >> (FRAC_PREC - 7);
      			texturecol(col, top, h, txc, th, buf, stonetex);	// vertical wall texture
#else
      			pixelcol89(col, top, h, buf, BUFPITCH);		// draw in the light plane
#endif      			
      			//pixelcol89(col, top, h, LCD_MEM, 30);
            //Zbuffer[col] = dst_x;
        }
        else if(yplot) {
            th = FIX_TO_INT(FIX_DIV(REFHEIGHT, dst_y));
      			h = (th < HEI) ? th : HEI;
      			top = HEI/2 - (h>>1);
#ifdef TEXTURE_WALLS
      			txc = tx_y >> (FRAC_PREC - 7);
      			texturecol(col, top, h, txc, th, buf, stonetex);  // horizontal wall texture
#else
      			pixelcol89(col, top, h, buf + BUFSIZE, BUFPITCH);	// draw in the dark plane
#endif      			
      			//pixelcol89(col, top, h, LCD_MEM, 30);
            //Zbuffer[col] = dst_y;
        }
        //else {
        //	// neither ray struck a wall--perhaps the maze is open	
        //    pixelcol89(col, 0, 5, buf, BUFPITCH);
        //}
    } // for(each column on screen)
}

/* - attempt 2 - still doesn't work
// Renders the world in 3D to the given buffer
void world_render3D(FIX cam_x, FIX cam_y, short angle, void *buf)
{
	short col, rangle, va, wa, th, h, top;
	short hyo, vxo;
	FIX hx, hy, hdx, hdy;
	FIX vx, vy, vdx, vdy;
	FIX hdist, vdist, htex, vtex, htcol, vtcol;
	FIX dist, tex, tcol;

    for(col=0; col < WID; col++)
	{
		// Normalize the viewing angle

		rangle = angle + col - WID/2;
		if(rangle >= ANG_360) rangle -= ANG_360;
		else if(rangle < 0) rangle += ANG_360;
		va = ABS(rangle - angle);

		// Setup grid intersection information

		if (rangle < ANG_90)
		{
			wa = rangle;
			
			hdx = cot_t[wa];
			hdy = FIX_1;
			hy = FIX_IPART(cam_y) + FIX_1;
			hx = cam_x + FIX_MUL(hy - cam_y, hdx);
			hyo = 0;

			vdx = FIX_1;
			vdy = tan_t[wa];
			vx = FIX_IPART(cam_x) + FIX_1;
			vy = cam_y + FIX_MUL(vx - cam_x, vdy);
			vxo = 0;
		}
		else if (rangle < ANG_180)
		{
			wa = ANG_180 - rangle;
			
			hdx = -cot_t[wa];
			hdy = FIX_1;
			hy = FIX_IPART(cam_y) + FIX_1;
			hx = cam_x + FIX_MUL(hy - cam_y, hdx);
			hyo = 0; 

			vdx = -FIX_1;
			vdy = tan_t[wa];
			vx = FIX_IPART(cam_x);
			vy = cam_y + FIX_MUL(cam_x - vx, vdy);
			vxo = -1; 
		}
		else if (rangle < ANG_270)
		{
			wa = rangle - ANG_180;

			hdx = -cot_t[wa];
			hdy = -FIX_1;
			hy = FIX_IPART(cam_y);
			hx = cam_x + FIX_MUL(cam_y - hy, hdx);
			hyo = -1;

			vdx = -FIX_1;
			vdy = -tan_t[wa];
			vx = FIX_IPART(cam_x);
			vy = cam_y + FIX_MUL(cam_x - vx, vdy);
			vxo = -1; 
		}
		else
		{
			wa = ANG_360 - rangle;
			
			hdx = cot_t[wa];
			hdy = -FIX_1;
			hy = FIX_IPART(cam_y);
			hx = cam_x + FIX_MUL(cam_y - hy, hdx);
			hyo = -1;

			vdx = FIX_1;
			vdy = -tan_t[wa];
			vx = FIX_IPART(cam_x) + FIX_1;
			vy = cam_y + FIX_MUL(vx - cam_x, vdy);
			vxo = 0;
		}

		// Find nearest horizontal intersection

		hdist = FIX_MAX;
		if(hx >= 0 && hx < m_fixwidth) 
		{
            while(0 == (htex = get(FIX_TO_INT(hx), FIX_TO_INT(hy) + hyo))) 
			{
                hx += hdx;
                hy += hdy;	// always +/- 1; so if the maze is closed, vertical bounds need not be explicitly checked
                if(hx < 0 || hx >= m_fixwidth)
                    goto next;
            }
            //hdist = FIX_DIV( FIX_MUL(ABS(hx - cam_x), cos_t[va]), cos_t[wa]);
            hdist = (ABS(hx - cam_x) * cos_t[va]) / cos_t[wa];	// tricky; the mul and div shifts cancel!
			htcol = FIX_FPART(hx);
        }
	
	next:		
		vdist = FIX_MAX;
		if(vy >= 0 && vy < m_fixheight) 
		{
            while(0 == (vtex = get(FIX_TO_INT(vx) + vxo, FIX_TO_INT(vy)))) 
			{
                vx += vdx;
                vy += vdy;	// always +/- 1; so if the maze is closed, horizontal bounds need not be explicitly checked
                if(vy < 0 || vy >= m_fixheight)
                    goto plot;
            }
            //vdist = FIX_DIV( FIX_MUL(ABS(vy - cam_y), cos_t[va]), sin_t[wa]);
            vdist = (ABS(vy - cam_y) * cos_t[va]) / sin_t[wa];	// tricky; the mul and div shifts cancel!
			vtcol = FIX_FPART(vy);
        }

        // determine which intersection was closer
	plot:
        if(hdist > 0 && hdist < vdist) {
            //th = FIX_TO_INT(FIX_DIV(REFHEIGHT, hdist));
			th = REFHEIGHT / hdist;	// again, the shifts cancel
            tex = htex;
            tcol = htcol;
            dist = hdist;
//            Zbuffer[col] = dst_x;
        }
        else if(vdist > 0 && vdist < FIX_MAX) {
            //th = FIX_TO_INT(FIX_DIV(REFHEIGHT, vdist));
			th = REFHEIGHT / vdist;
            tex = vtex;
            tcol = vtcol;
            dist = vdist;
//            Zbuffer[col] = dst_y;
        }
		else {
			plotpixel(col, 110, 255, 0, 0);
			plotpixel(col, 112, 255, 0, 0);
			plotpixel(col, 114, 255, 0, 0);
			return;	// neither ray struck a wall--this won't happen if the maze is closed
		}

        plotpixel(col, 110, hdist, hdist, hdist);
		plotpixel(col, 112, vdist, vdist, vdist);
		plotpixel(col, 114, th, th, th);
		
		h = (th < HEI) ? th : HEI;

        // plot the column!!!
        top = HEI/2 - (h>>1);
		pixelcol89(col, top, h, buf, WID / 8);

	}	// for each col
}
*/
